#ifndef _CNF_STMT_H
#define _CNF_STMT_H

#include "ipp.h"
#include "SATSolver.h"

#include <hash_map>
#include <vector>
#include <string>
#include <list>

class CNF_STMT {
public:
  CNF_STMT() : my_contradiction(-1) {}
  CNF_STMT(fact_list stmt) : my_contradiction(-1) { set(stmt); }

  void add(fact_list clause);
  void set(fact_list stmt);

  void compact();

  bool solve(hash_map<string, bool> & soln, hash_map<string, double> & vsoln,
		   bool & timedOut);
  
  void prettyPrint(ostream & out = cout);
  void print(ostream & outCNF, ostream & outMap);

  void readSolution(istream & inSol, istream & inMap,
				hash_map<string, bool> & soln,
				hash_map<string, double> & vsoln);

private:
  int size();

  void set(VAR_OCC lit);

  void add(token_list clause);

  void add_token(string & str);
  int find_token(string str);
  string find_token(int i);
  
  void add_var(string var);
  int find_var(string var);
  string find_var(int var);

  void add_constraint(string str);

  void add(CNF_CLAUSE c);

  VAR_OCC find_pure_lit();

  void streamline_clause(CNF_CLAUSE & c);
  void unit_propagate();
  void subsume();
  void eliminate_pure_literals();
  void renumber();

  bool subsumes(CNF_CLAUSE small, CNF_CLAUSE big);

  pair< hash_map<string, int>, vector<string> > my_map;
  pair< hash_map<string, int>, vector<string> > my_vmap;
  vector< CNF_CLAUSE > my_clauses;
  vector< list<pair<int, bool> > > my_vars;
  hash_map< int, string > my_constraints;
  hash_map< int, bool > unit_vars;
  int my_contradiction;

  friend class Clause;
};


#endif
