/* (C) Copyright 1997 Albert Ludwigs University Freiburg
 *     Institute of Computer Science
 *
 * All rights reserved. Use of this software is permitted for 
 * non-commercial research purposes, and it may be copied only 
 * for that use.  All copies must include this copyright message.
 * This software is made available AS IS, and neither the authors
 * nor the  Albert Ludwigs University Freiburg make any warranty
 * about the software or its performance. 
 */

/* 	$Id: eq_preproc.c,v 1.10 1998/10/28 23:40:36 wolf Exp $	 */

#ifndef lint
static char vcid[] = "$Id: eq_preproc.c,v 1.10 1998/10/28 23:40:36 wolf Exp $";
#endif /* lint */


#include"ipp.h"/* defines, data structures, fn prototypes, global variables */

#include <assert.h>

/* function prototypes */
void preproc_op_hooks( op_list op );
void preprocess_fact_list( fact_list * fl );
void get_normal_form_cond( op_list op );
fact_list rec_eliminate_quantifiers( fact_list f, fact_list *end );
fact_list rec_rewrite_in_cnf( fact_list* cond );
fact_list multiply_clauses( fact_list old1, fact_list old2 );

/* checks if a token_list consists of exactly two equal strings */
BOOLEAN args_eq( token_list two_args )
{
  if (!two_args || !two_args->item || 
	 !two_args->next || !two_args->next->item ||
	 two_args->next->next)
    return FALSE;

  return (strcmp(two_args->item, two_args->next->item) == SAME);
}

/* make copy of a token_list and save the end of that
   list (...to attach further elements) */
token_list copy_tl_return_end( token_list tl, token_list *end )
{
  token_list t, start;
  
  *end = NULL;
  start = t = new_token_list();
  for( ; tl; tl = tl->next )
    {
      t->next = new_token_list();
      t = t->next;
      t->item = tl->item;
      if ( !tl->next )
	{
	  *end = t;
	}
    }
  return start->next;

}




/**********************************************************************/
/**********************************************************************/
/**********************************************************************/


void preprocess_domain_vars()
{
  fact_list current = domain_vars_list, last = NULL;

  while (current) {
    /* init numbers would go in the third slot... check for them: */
    if (current->item->next->next != NULL) {
	 /* FLUENT and FLOAT must be handled differently! */
	 if (strcmp(current->item->next->item, FLOAT_STR) == SAME) {
	   /* FLOAT - replace with number, remove from domain_vars_list,
		 give warning if ever set */
	   token old = current->item->item;
	   token new_tok = current->item->next->next->item;
	   fact_list fl = current;

	   if (last == NULL) {
		domain_vars_list = current->next;
		current = current->next;
		fl->next = NULL;
	   }
	   else {
		last->next = current->next;
		current = current->next;
		fl->next = NULL;
	   }

	   global_replace_cons_var(old, new_tok);
	   free_complete_fact_list(fl);
	 }
	 else if (strcmp(current->item->next->item, FLUENT_STR) == SAME) {
	   /* FLUENT - add constraint to initial conditions, remove number
		 from dom_vars_list */
	   token_list constraint, tl;
	   fact_list fl;

	   constraint = tl = new_token_list();
	   tl->item = new_token(strlen(EFFECT_STR) + 1);
	   strcpy(tl->item, EFFECT_STR);

	   tl->next = new_token_list();
	   tl = tl->next;
	   tl->item = new_token(strlen(SET_STR) + 1);
	   strcpy(tl->item, SET_STR);

	   tl->next = new_token_list();
	   tl = tl->next;
	   tl->item = new_token(strlen(current->item->item) + 1);
	   strcpy(tl->item, current->item->item);

	   tl->next = new_token_list();
	   tl = tl->next;
	   tl->item = new_token(strlen("=") + 1);
	   strcpy(tl->item, "=");

	   tl->next = current->item->next->next;

	   current->item->next->next = NULL;

	   fl = new_fact_list();
	   fl->item = constraint;
	   fl->next = orig_initial_facts;
	   orig_initial_facts = fl;

	   last = current;
	   current = current->next;
	 }
	 else {
	   /* ERROR! there are only these two types. */
	   assert (strcmp(current->item->next->item, FLUENT_STR) == SAME ||
			 strcmp(current->item->next->item, FLOAT_STR) == SAME);
	 }
    }
    else {
	 last = current;
	 current = current->next;
    }
  }
}

void preprocess_pl1_facts( void )
{
  /* step through all operator pre- and effect conditions and
     transform these condtions into a format which is processable
     for ipp. */
  op_list i_op;

  for (i_op = loaded_ops; i_op; i_op = i_op->next) {
    preproc_op_hooks( i_op );
  }

  /* now handle the goals */
  /* Wolf - I changed this to handle goals as fact_list rather than action */
  preprocess_goals();
}


void preproc_op_hooks( op_list op )
{
  if ( !op )
    return;
  
  /* then check the preconditions. attention: here new operators
     are built, the old one may be destroyed */
  /* Wolf - since everything is now in the same format, I process
	both "preconds" and "effects" here. */
  get_normal_form_cond( op ); 
}

void preprocess_goals()
{
  preprocess_fact_list( &raw_goal_facts );
}

void preprocess_fact_list(fact_list * fl)
{
  if ( !fl || !(*fl) ) 
    {
	 return;
    }


  *fl = transform_to_atomic_negation( *fl, FALSE );
  *fl = eliminate_quantifiers( *fl );
  *fl = rewrite_in_cnf( fl );
}

void get_normal_form_cond( op_list op )
{
  if ( !op )
    return;

  preprocess_fact_list( &op->hooks );
}


fact_list transform_to_atomic_negation( fact_list f, BOOLEAN neg )
{
  int c;
  fact_list result;
  token t;

  if ( !f )
    return NULL;

  /* find out the first item in f */
  c = get_adl_token( f );

  result = f;
  switch ( c )
    {
    case LIT_CONST:
      /* it'a literal. if neg == TRUE, it must be negated */
      if ( neg )
	{
	  token_list tl = f->item;
	  if ( tl->item[0] == '!' )
	    {
	      t = new_token( strlen(tl->item)-1+1 );
	      strcpy( t, (tl->item)+1 );
	      my_free( tl->item );
	      tl->item = t;
	    }
	  else
	    {
	      t = new_token( strlen(tl->item)+1+1 );
	      *t = '!';
	      strcpy( t+1, tl->item );
	      my_free( tl->item );
	      tl->item = t;
	    }
	}
      result->next = transform_to_atomic_negation( f->next, neg );
      break;
    case VAR_CONST: 
    case LBRACK_CONST: 
    case RBRACK_CONST:
      /* do nothing - BTW: VAR occurs in forall/exist clauses - Wolf */
      result->next = transform_to_atomic_negation( f->next, neg );
      break;
    case AND_CONST:
      if ( neg )
	{
	  result = make_adl_fact( OR_CONST );
	  result->next = transform_to_atomic_negation(f->next,neg );
	  f->next = NULL;
	  free_complete_fact_list( f );
	}
      else
	{
	  result->next = transform_to_atomic_negation( f->next, neg );
	}
      break;
    case OR_CONST:
      if ( neg )
	{
	  result = make_adl_fact( AND_CONST );
	  result->next = transform_to_atomic_negation(f->next,neg );
	  f->next = NULL;
	  free_complete_fact_list( f );
	}
      else
	{
	  result->next = transform_to_atomic_negation( f->next, neg );
	}
      break;
    case EXISTS_CONST:
      if ( neg )
	{
	  result = make_adl_fact( FORALL_CONST );
	  result->next = transform_to_atomic_negation(f->next,neg );
	  f->next = NULL;
	  free_complete_fact_list( f );
	}
      else
	{
	  result->next = transform_to_atomic_negation( f->next, neg );
	}
      break;
    case FORALL_CONST:
      if ( neg )
	{
	  result = make_adl_fact( EXISTS_CONST );
	  result->next = transform_to_atomic_negation(f->next,neg );
	  f->next = NULL;
	  free_complete_fact_list( f );
	}
      else
	{
	  result->next = transform_to_atomic_negation( f->next, neg );
	}
      break;
    case NOT_CONST:
      /* drop NOT here and move it into the literals */
      result = transform_to_atomic_negation( f->next, 
					     neg ? FALSE : TRUE );
      f->next = NULL;
      free_complete_fact_list( f );
      break;
    case ENDNOT_CONST:
      /* NOT scope is over */
      result = transform_to_atomic_negation( f->next, 
					     neg ? FALSE : TRUE ); 
      f->next = NULL;
      free_complete_fact_list( f );
      break;
    default:
      fprintf(stderr, "error: unknown pl1 constant detected: %d\n", c );
      OUTPUT_FILE;
      exit( 1 );
      break;
    }

  return result;
}

fact_list eliminate_quantifiers( fact_list f )
{
  fact_list result = NULL;
  fact_list end = f;
  fact_list temp;
  fact_list rend = NULL;

  if (f == NULL) 
    return NULL;

  do {
    temp = rec_eliminate_quantifiers( f, &end );

    if (!result) {
	 result = temp;
	 rend = end;
    }
    else {
	 rend->next = temp;
	 rend = end;
    }

    f = end->next;
  } while (f != NULL);

  return result;
}

/* f is a first order logic expression, *end will hold the last
   processed item after this function was run, i.e. (*end)->next
   will point to the next item that was not processed yet */
fact_list rec_eliminate_quantifiers( fact_list f, fact_list *end )
{
  int c;
  fact_list result, vars, v, expr, tmp, following, *vinst_fl;
  token_list ol, o;

  *end = f;

  if ( !f )
    return NULL;

  /* find out which token comes next */
  c = get_adl_token( f );

  switch ( c )
    {
    
    case LIT_CONST:
      /* if the actual fact is a simple literal => return it */
      return f;
      break;

    case AND_CONST:
    case OR_CONST:
      /* if it's AND or OR, that's okay, but eliminate quants in the 
	    arguments to AND... */
      while ( get_adl_token((*end)->next) != RBRACK_CONST )
	   {
		assert(*end);
		(*end)->next = rec_eliminate_quantifiers( (*end)->next, &tmp );
		*end = tmp;
	   }
      /* now point *end to the closing AND bracket */
	 *end = (*end)->next;
      return f;

    case FORALL_CONST:
    case EXISTS_CONST:
      /* we have to get the variables bound by the quant, then get
	    the (already quantifier-free) expression and copy it
	    for each possible instantiation of the variable. Finally
	    build the conjuction over all of the new expressions */
      if ( c == FORALL_CONST )
	   {
		result = make_adl_fact( AND_CONST );
	   }
      else
	   {
		result = make_adl_fact( OR_CONST );
	   }
      *end = f->next; /* skip quantifier and left bracket */
	 f->next = NULL;
	 free_complete_fact_list(f);
      vars = get_variables( end );
      /* after getting the variables, end points to the expression */
      expr = rec_eliminate_quantifiers( *end, &tmp );
      following = tmp->next;
      tmp->next = NULL;
      /* end points to OR/AND - (*end)->next will be the point where
	    all possible instatiations will be appended */
      for ( v = vars; v; v = v->next )
	   {
		if ( !( ol = type_already_known( v->item->next->item, 
								   global_object_fl ) ) )
		  {
		    fprintf(stderr, "error: type %s for %s not specified.\n",
				  v->item->next->item,  v->item->item );
		    OUTPUT_FILE;
		    exit( 1 );
		  }
		vinst_fl = &(result->next);
		for ( o = ol; o; o = o->next )
		  { /* instatiate v with o */
		    *vinst_fl = copy_complete_fact_list( expr, &tmp );
		    replace_in_fl( v->item->item, o->item, *vinst_fl );
		    vinst_fl = &(tmp->next);
		  }
		/* now variable v is instatiated, the list of facts was
		   blown up... but our work is yet to start: instantiate
		   the others vars in new new list */
		expr = result->next;
		*end = tmp;
	   }


	 /* no need to close AND/OR; just use RBRACK already there from
	    FORALL/EXISTS */
      (*end)->next = following;
      *end = (*end)->next;
      return result;
      break;
    default: 
      fprintf(stderr, "\nstrange first order object %d!\n", c );
      OUTPUT_FILE;
      exit( 1 );
    }
}

fact_list rewrite_in_cnf( fact_list* cond )
{
  fact_list result = NULL, curr, end = NULL;
  while (*cond) {
    curr = rec_rewrite_in_cnf(cond);
    if (result) {
	 end->next = curr;
    }
    else {
	 result = end = curr;
    }
    for ( ; end->next; end = end->next)
	 ;
  }

  return result;
}

/* Wolf - this is based on rec_remove_quants_from_precond, but instead of achieving
   DNF (yuck!) this gets CNF (yeah!) */
fact_list rec_rewrite_in_cnf( fact_list* cond )

{
  fact_list result = NULL, fl, end;
  fact_list tmp;
  int c;
  
  if ( !(*cond) )
    {
      /* recursion should be ended by the last bracket */
      fprintf(stderr, "error: precondition is no correct pl1 expression\n" );
      OUTPUT_FILE;
      exit( 1 );
    }

  c = get_adl_token( *cond );

  switch ( c )
    {
    case RBRACK_CONST:
      fprintf(stderr, "\nerror: parsing adl expression in preconds. debug me.\n" );
      OUTPUT_FILE;
      exit ( 1 );
      break;
    case LIT_CONST:
	 /* make this into a clause for consistency (everything above LIT
	  * level can assume clauses) */
	 fl = make_adl_fact(OR_CONST);
	 /* throw in one literal (will build up list in OR and AND cases) */
	 fl->next = *cond;

      /* no copying of *cond necessary here */
      *cond = (*cond)->next;

	 fl->next->next = make_adl_fact(RBRACK_CONST);
      result = fl;
      break;
    /* Wolf - I took the easy way out and rewrote AND from the old function as OR here
	  and vice versa */
    case OR_CONST:
      /* skip OR_CONST */
	 tmp = *cond;
      *cond = (*cond)->next;
	 my_free(tmp);

	 result = NULL;
      while (RBRACK_CONST != get_adl_token( *cond ))
	   { /* still OR */
		fl = rec_rewrite_in_cnf( cond );
		if ( !fl )
		  { 
		    fprintf(stderr,"warning: OR clause disjunct came back NULL.\n");
		    continue;
		  }
		if ( !(*cond) )
		  {
		    fprintf(stderr, "error: OR clause must be terminated with ')'\n" );
		    OUTPUT_FILE;
		    exit( 1 );
		  }
		/* build the cartesian product of the elements (elementary
		   operation is disjunction) */
		if (result) 
		  result = multiply_clauses( result, fl );
		else
		  result = fl;
	   }
      /* delete RBRACK_CONST */
	 tmp = *cond;
      *cond = (*cond)->next;
	 my_free(tmp);
      break;
    case AND_CONST:
      /* skip AND_CONST */
      tmp = *cond;
      *cond = (*cond)->next;
      my_free( tmp );
      /* get one of the conjunctive effects after another */
      result = end = NULL;
      while ( RBRACK_CONST != get_adl_token( *cond ) )
	   { /* still AND */
		fl = rec_rewrite_in_cnf( cond );
		if ( !(*cond) )
		  {
		    fprintf(stderr, 
				  "error: AND clause must be terminated with ')'\n" );
		    OUTPUT_FILE;
		    exit( 1 );
		  }
		if (!fl) {
		  fprintf(stderr,
				"warning: AND clause conjunct came back NULL.\n");
		  continue;
		}

		if ( !result )
		  result = end = fl;
		else
		  end->next = fl;

		if ( end )
		  for ( ; end->next; end = end->next )
		    ;
	   }
      /* skip RBRACK_CONST */
      tmp = *cond;
      *cond = (*cond)->next;
      my_free( tmp );
      break;
    case EXISTS_CONST:
    case FORALL_CONST:
      fprintf(stderr, 
		    "\nerror: quantifier was not eliminated during preprocessing." );
      OUTPUT_FILE;
      exit( 1 );
      break;

    }
  return result;
}

fact_list get_variables( fact_list* last )
{
  fact_list f = *last, result = (*last)->next, tmp;
  int c = 0;
  
  /* first fact in *last is a left bracket */
  for ( *last = f; (*last)->next; *last = (*last)->next )
    {
      c = get_adl_token((*last)->next);
      if ( RBRACK_CONST == c )
	break;
      if ( VAR_CONST != c )
	{
	  fprintf(stderr, "\nerror: non-variable (%s) in parameter list: ",
		adl_string[c] );  
	  fprintf(stderr, "'%s'\n", (*last)->next->item->item );
	  OUTPUT_FILE;
	  exit( 1 );
	}
    }
  if ( !((*last)->next) )
    {
      fprintf(stderr, "\nerror: variable list doesn't end with ')'\n" );
      fprintf(stderr, "\nbut ends with %s", adl_string[c] );
      fprintf(stderr, ": %s...\n", (*last)->item->item );
      print_factlist( result );
      OUTPUT_FILE;
      exit( 1 );
    }
  /* up to this point we have a list of variables */
  tmp = (*last)->next;
  (*last)->next = NULL; 
  *last = tmp;
  /* skip bracket */
  tmp = *last;
  *last = (*last)->next;
  my_free( tmp );
  return result;
}

void replace_in_cons_params( token old, token new_tok, token_list source )
{
  while (source) {
    if (source->item && source->item[0] == '?') {
	 char * look_start = strchr(source->item, '[');
	 char * str_start;

	 if (look_start)
	   while ((str_start = strstr(look_start, old)) &&
			(*(str_start - 1) == '[' || *(str_start - 1) == ',') &&
			(*(str_start + strlen(old)) == ']' ||
			 *(str_start + strlen(old)) == ',')) {
		token t = new_token(strlen(source->item) - strlen(old) + 
						strlen(new_tok) + 1);
		
		*str_start = '\0';
		sprintf(t, "%s%s%s", source->item, new_tok, str_start + strlen(old));
		my_free(source->item);
		source->item = t;
	   }
    }
    
    source = source->next;
  }
}

void replace_in_fl( token old, token new_tok, fact_list source )
{
  fact_list f;
  token_list t;
  token n;

  int len = strlen( new_tok ) + 1;
  for ( f = source; f; f = f->next )
    {
	 if (is_constraint_form(f)) {
	   replace_in_cons_params(old, new_tok, f->item);
	 }
      for ( t = f->item; t; t = t->next )
	   {
		if ( strcmp( t->item, old ) == SAME )
		  { 		  
		    n = new_token( len );
		    strcpy( n, new_tok );
		    my_free( t->item );
		    t->item = n;
		  }
	   }
    }
}



fact_list multiply_clauses( fact_list old1, fact_list old2 )
{
  fact_list result = NULL, tmp1, tmp2, end = NULL, new_end;
  fact_list dummy;

  /* start each iteration of the loop with tmp1 at the start of a clause */
  for ( tmp1 = old1; tmp1; tmp1 = tmp1->next )
    {
	 assert(get_adl_token(tmp1) == OR_CONST);
	 tmp1 = tmp1->next;

	 /* start each iteration of the loop with tmp2 at the start of a clause */
      for ( tmp2 = old2; tmp2; tmp2 = tmp2->next )
	   {
		assert(get_adl_token(tmp2) == OR_CONST);
		tmp2 = tmp2->next;

		dummy = copy_complete_fact_clause(tmp1, &new_end); 
		        /* copy old clause to RBrack */
		        /* does not actually copy    */
                  /* bracket                   */

		if ( !result ) {
		  result = make_adl_fact(OR_CONST);
		  result->next = dummy;
		}
		else {
		  end->next = make_adl_fact(OR_CONST);
		  end->next->next = dummy;
		}
		
		end = new_end;

		dummy = copy_complete_fact_clause(tmp2, &new_end); 
                                                    /* copy new clause */

		end->next = dummy;
		end = new_end;
		end->next = make_adl_fact(RBRACK_CONST);
		end = end->next;

		/* Advance tmp2 to the end of the clause */
		for ( ; get_adl_token(tmp2) != RBRACK_CONST; tmp2 = tmp2->next )
		  if (tmp2->next == NULL) {
		    printf("error: malformed clause in multiply_clauses.\n");
		    OUTPUT_FILE;
		    exit ( 1 );
		  }
	   }

	 /* Advance tmp1 to the end of the clause */
	 for ( ; get_adl_token(tmp1) != RBRACK_CONST; tmp1 = tmp1->next )
	   if (tmp1->next == NULL) {
		printf("error: malformed clause in multiply_clauses.\n");
		OUTPUT_FILE;
		exit ( 1 );
	   }
    }
  
  free_complete_fact_list( old1 );
  free_complete_fact_list( old2 );
  
  return result;
}
