;; logistics domain
;;
;; logistics-strips: pure strips, no types, no equality
;;    Compatible with problem instances for the domain01.pddl 
;;    from the PDDL 1.3 distribution.
;;    Fixes a few missing parameter predicates and
;;    has unload truck and plane as separate operators
;;    (as in original formulation).
;;    12 June 1998, Bart Selman & Henry Kautz
;;
;; Modified to work with lpsat as a zero metric test case.

(define (domain logistics-strips)
  (:requirements :strips) 
  (:types LOCATION CITY MOVABLE - OBJECT OBJ VEHICLE - MOVABLE
		AIRPORT - LOCATION TRUCK AIRPLANE - VEHICLE)
  (:predicates
   (at ?obj - MOVABLE ?loc - LOCATION)
   (in ?obj - OBJ ?obj - VEHICLE)
   (in-city ?obj - LOCATION ?city - CITY))
 
(:action LOAD-TRUCK
  :parameters
   (?obj - OBJ
    ?truck - TRUCK
    ?loc - LOCATION)
  :precondition
   (and (at ?truck ?loc) (at ?obj ?loc))
  :effect
   (and (not (at ?obj ?loc)) (in ?obj ?truck)))

(:action LOAD-AIRPLANE
  :parameters
   (?obj - OBJ
    ?airplane - AIRPLANE
    ?loc - AIRPORT)
  :precondition
   (and (at ?obj ?loc) (at ?airplane ?loc))
  :effect
   (and (not (at ?obj ?loc)) (in ?obj ?airplane)))

(:action UNLOAD-TRUCK
  :parameters
   (?obj - OBJ
    ?truck - TRUCK
    ?loc - LOCATION)
  :precondition
   (and (at ?truck ?loc) (in ?obj ?truck))
  :effect
   (and (not (in ?obj ?truck)) (at ?obj ?loc)))

(:action UNLOAD-AIRPLANE
  :parameters
   (?obj - OBJ
    ?airplane - AIRPLANE
    ?loc - AIRPORT)
  :precondition
   (and (in ?obj ?airplane) (at ?airplane ?loc))
  :effect
   (and (not (in ?obj ?airplane)) (at ?obj ?loc)))

(:action DRIVE-TRUCK
  :parameters
   (?truck - TRUCK
    ?loc-from - LOCATION
    ?loc-to - LOCATION
    ?city - CITY)
  :precondition
   (and (at ?truck ?loc-from)
	   (in-city ?loc-from ?city)
	   (in-city ?loc-to ?city))
  :effect
   (and (not (at ?truck ?loc-from)) (at ?truck ?loc-to)))

(:action FLY-AIRPLANE
  :parameters
   (?airplane - AIRPLANE
    ?loc-from - AIRPORT
    ?loc-to - AIRPORT)
  :precondition
   (and (at ?airplane ?loc-from))
  :effect
   (and (not (at ?airplane ?loc-from)) (at ?airplane ?loc-to)))
)
