;; logistics domain
;;
;; logistics-strips: pure strips, no types, no equality
;;    Compatible with problem instances for the domain01.pddl 
;;    from the PDDL 1.3 distribution.
;;    Fixes a few missing parameter predicates and
;;    has unload truck and plane as separate operators
;;    (as in original formulation).
;;    12 June 1998, Bart Selman & Henry Kautz
;;
;;    13 Oct 1998, Steve Wolfman - modified to metricize (include 
;;                                 metric effects/preconditions/parms)
;;
;; Question: should airplanes have a separate (at ?a - AIRPLANE ?l - AIRPORT)?
;;
;; Note: changed so that planes can load and unload only at airports.
;;       only possible problem I can think of with this is if AIRPLANEs
;;       can be OBJs (and moved around be trucks... cool thought!)
;;

(define (domain logistics-metric)
  (:requirements :strips) 
  (:types LOCATION CITY MOVABLE - OBJECT OBJ VEHICLE - MOVABLE
		AIRPORT - LOCATION TRUCK AIRPLANE - VEHICLE)
  (:predicates 	(at ?obj - MOVABLE ?loc - LOCATION)
				(in ?obj - OBJ ?obj - VEHICLE)
				(in-city ?loc - LOCATION ?city - CITY)
				(depot ?loc - LOCATION)
				)
  (:action LOAD-AIRPLANE
		 :parameters
		 (?obj - OBJ
		  ?airplane - AIRPLANE
		  ?loc - AIRPORT)
		 :precondition
		 (and (at ?airplane ?loc) (at ?obj ?loc))
		 :effect
		 (and (not (at ?obj ?loc)) (in ?obj ?airplane)))
  
  (:action REFUEL
		 :parameters
		 (?veh - VEHICLE
		  ?loc - LOCATION)
		 :precondition
		 (and (at ?veh ?loc) (depot ?loc))
		 :effect
		 (and (set ?fuel[?veh] ?capacity[?veh])))

  (:action FLY-AIRPLANE
		 :parameters
		 (?airplane - AIRPLANE
		  ?loc-from - AIRPORT
		  ?loc-to - AIRPORT)
		 :precondition
		 (and (at ?airplane ?loc-from)
			 (test ?fuel[?airplane] >= ?airplane_trip + 
				           ?dist[?loc-from,?loc-to] / ?plane_mpg))
		 :effect
		 (and (not (at ?airplane ?loc-from)) (at ?airplane ?loc-to)
			 (set ?fuel[?airplane] ?fuel[?airplane] - ?airplane_trip - 
				       ?dist[?loc-from,?loc-to] / ?plane_mpg)))
  )
