/* (C) Copyright 1997 Albert Ludwigs University Freiburg
 *     Institute of Computer Science
 *
 * All rights reserved. Use of this software is permitted for 
 * non-commercial research purposes, and it may be copied only 
 * for that use.  All copies must include this copyright message.
 * This software is made available AS IS, and neither the authors
 * nor the  Albert Ludwigs University Freiburg make any warranty
 * about the software or its performance. 
 */

#include"ipp.h"/* defines, data structures, fn prototypes, global variables */

int unique_fluent_id = 1;

void instantiate_list( op_list ops, 
				   fact_list constants )
{

  op_list op_i;

  operators = NULL;

  for( op_i = ops; op_i; op_i = op_i->next )
    instantiate( *op_i, constants );
}


/* instantiate ONE operator */
void instantiate( op_list_elt op, 
			   fact_list constants )

{
  instantiate_first_parameter( op, constants );
}

/* instantiates first parameter of given op with a possible
 * constant from given list and calls itself recursively
 * with now partially instantiated op, one parameter( the first )
 * and one ( the already used ) constant less.
 *
 * all_constants always contains what its name says, needed later by
 * delete_ALL_quantifying
 */
void instantiate_first_parameter( op_list_elt op,
                                  fact_list constants )
{

  operator_list oper;
  fact_list i_constant, i_param;
  token_list parameter;
  op_list_elt help_op;
  fact_list help_constants, i_fact;

  for ( i_param=op.params; i_param; i_param=i_param->next ) {
    for ( i_fact=op.params_objects; i_fact; i_fact=i_fact->next ) 
      if ( strcmp( i_param->item->item, i_fact->item->item ) == SAME ) break;
    if ( !i_fact ) break;
  } 
  if ( !i_param ) {
    /*op is instantiated; store in global list*/

    /* before switching to operator, rename fluent params to be 
	  unique */
    rename_fluent_params(&op, unique_fluent_id++);

    /* also rename influence objects if influences are on */
    if (bInfluences)
	 rename_influence_objects(op.hooks, unique_fluent_id);

    oper = new_operator_list( make_name( op ) );
    /* check if all variables are initialized */
    check_instantiation(op.hooks, op.name);

    /* Wolf - eliminate trivial eq/!eq (should really elim trivial
	         timeless stuff!!!): */
    eliminate_timeless(&op.hooks);
    oper->hooks = op.hooks;
    
    oper->objects = make_objects( op );
    oper->next = operators;
    operators = oper;
  } else {
    parameter = i_param->item;
    /* step through possible constants... */
    for ( i_constant = constants; i_constant; i_constant = i_constant->next ) {
      /* step through constants that were already used */
      for ( i_fact=op.params_objects; i_fact; i_fact=i_fact->next ) 
	   { 
		if ( strcmp( i_constant->item->item, i_fact->item->next->item )
			== SAME ) break;
	   }
      /* if a constant was already used try another one, continue loop */
      if ( i_fact ) continue;

      /* if type of parameter == type of i_constant */
      if (strcmp(parameter->next->item, i_constant->item->next->item)==SAME) 
	   {
		/* copy op into help_op, replacing parameter with i_constant
		 * and removing parameter from op.params list
		 */
		help_op = copy_replace_op( op,parameter->item,
							  i_constant->item->item);
	  
		/* copy constants into help_constants, removing i_constant */
		help_constants = copy_remove( constants, i_constant );
		/* now instantiate next parameter */
		instantiate_first_parameter( help_op, help_constants );
	   }
    }
  }

}

/* prints list of instantiated operators to stdout
 */
void print_operators( operator_list operators )

{

  operator_list i_operator;

  for ( i_operator = operators; i_operator; i_operator = i_operator->next ) {
    printf( "%s\n", i_operator->name );
    printf( "hooks:\n" );
    print_factlist(i_operator->hooks);

    printf( "\n" );
  }
}
