/* (C) Copyright 1997 Albert Ludwigs University Freiburg
 *     Institute of Computer Science
 *
 * All rights reserved. Use of this software is permitted for 
 * non-commercial research purposes, and it may be copied only 
 * for that use.  All copies must include this copyright message.
 * This software is made available AS IS, and neither the authors
 * nor the  Albert Ludwigs University Freiburg make any warranty
 * about the software or its performance. 
 *
 * Modified by Steve Wolfman 1999. Modified parser to read constraints
 * and altered parser for use with RelSAT system.
 */

#ifndef __IPP_H
#define __IPP_H

/*
 * defines,
 * data structures,
 * function prototypes and
 * global variables
 *
 *  for extended graphplan algorithm
 */


#include<stdlib.h>
#include<stdio.h>
#include<strings.h>
#include<ctype.h>
#include<sys/types.h>
#include<sys/times.h>
#include<hash_map>
#include<string>
#include<list>

#include "shared.h"

/*
 * defines
 */

/* Wolf - moved these to ipp.h from scan-ops.y */
#define STANDARD_TYPE "OBJECT\0"

/* Wolf - useful names for types and predicates */
#define FLOAT_STR "FLOAT\0"
#define FLUENT_STR "FLUENT\0"


/* technical defines */
#define MAX_LENGTH 1024/* maximal string length */
#define BOOLEAN unsigned short int/* you should know about that one */
#define TRUE 1/* this is true */
#define FALSE 0/* and this is not */
#define SAME 0/* for strcmp */
#define BIGPRIME 8000977/* prime for hashing */
#define NOOP "noop"/* der blanke wahnsinn */
#define INIT 1/* for initialising static variables, */
#define EXEC 0/* i.e., better readability of fn calls */
#define SEARCH 1/* readability */
#define COMPLETE 0/* of fn call( in search ) */
#define S_HSIZE 4/* size of tables in sub tree branches */
#define S_HASH 3/* for index calculation */
#define NO_QUANT 0
#define ALL_QUANT 1
#define EX_QUANT 2
#define QUANTIFIERS "-AE" /* no, universal, existential */
#define ANY_PRED 0
#define EQ 1
#define EQ_STR "eq"
#define NOT_EQ 2
#define NOT_EQ_STR "!eq" /* this one MUST be NOT_STR + EQ_STR ! */
#define NOT_STR "!"
#define NOT_PRED '!'
#define NOT_EQ_PRED "!eq"
#define NO_SOLUTION "NO SOLUTION\n"
#define NO_MEMORY "\nipp:  SORRY, I RAN OUT OF MEMORY!\n"
/* the following constants are for pl1 expression preprocessing */
#define LIT_CONST      1
#define AND_CONST      2
#define OR_CONST       3
#define IMPLY_CONST    4
#define NOT_CONST      5
#define EXISTS_CONST   6 
#define FORALL_CONST   7
#define LBRACK_CONST   8
#define RBRACK_CONST   9
#define VAR_CONST     10
#define ENDNOT_CONST  11
extern char *adl_string[];
#define HIDDEN_STR "#"
#define AXIOM_STR "AXIOM"
#define NAME_STR "name\0"
#define VARIABLE_STR "variable\0"
#define STANDARD_TYPE "OBJECT\0"
#define GOAL_OP_STR "#REACHGOAL"
#define GOAL_REACHED "#GOALREACHED"
#define EITHER_STR "EITHER"

/* Argument related info */
#define DETAIL_LEVEL_STRING "Set printing detail level:
\t\t0: Print only the plan solution.
\t\t1: Print high level info (run time, file names, ...)
\t\t2: Print statistics (number of constraint 
\t\t\tadds/deletes, number of choice points, ...)
\t\t3: Print high level debugging info (broad trace
\t\t\tof program progress)
\t\t4: Print all available debugging info
\t\t5: Print the word \"FLIG\" 1000 times, and quit"

#define DETAIL_NONE 0
#define DETAIL_MINIMAL 1
#define DETAIL_STATS 2
#define DETAIL_MINIMAL_DEBUG 3
#define DETAIL_ALL_DEBUG 4
#define DETAIL_FLIG 5

#define COMPACT_LEVEL_STRING "Set initial compact level:
\t\t0: No initial compaction
\t\t1: Initial unit propagate
\t\t2: Unit and pure
\t\t3: Unit, pure, and subsume"

#define COMPACT_NONE 0
#define COMPACT_UNIT 1
#define COMPACT_SUBSUME 3
#define COMPACT_PURE 2
#define COMPACT_ALL 3

#define CACHE_POLICY_STRING "Set cache policy (use with cache-depth option):
\t\t0: No caching
\t\t1: Flush based on cache size
\t\t2: Flush based on number of branches"

#define NO_CACHE 0
#define CACHE_SIZE 1
#define CACHE_BRANCH 2

#define OUTPUT_FILE log_end(TOTAL_TIME); \
  cerr << NO_SOLUTION << endl; \
  wrap_up();
/* Macro for NULL pointer check after mallocs/callocs */
#define CHECK_MEMORY(ptr) if (!(ptr)) {\
  cerr << NO_MEMORY;\
  OUTPUT_FILE;\
  exit(1);}

/* Wolf - defines for handling: 1) constraints and 2) merged precond/effects */
#define PRECOND_STR "#<PRECOND>\0"
#define EFFECT_STR "#<EFFECT>\0"
#define NOT_PRECOND_STR "!#<PRECOND>\0"
#define NOT_EFFECT_STR "!#<EFFECT>\0"

/* Wolf - defines for time logging strings */
#define TOTAL_TIME "Total time"

/*
 * data structures
 */


/* obviously a string */
typedef char *token;

/* string list */
typedef struct TOKENLIST {

  token item;
  struct TOKENLIST *next;

} *token_list, token_list_elt;


/* list of string list */
typedef struct FACTLIST {

  token_list item;
  struct FACTLIST *next;

} *fact_list, fact_list_elt;


/* uninstantiated operator, read from inputfile <name>.ops
 * into this structure
 */
typedef struct OP {

  char *name;

  fact_list params;
  token_list fluent_params; /* to store the constrainable parameters */
  fact_list hooks; /* Wolf - preconditions (effects before) and 
				  effects (preconditions after :) */

  fact_list params_objects;

  int number_of_real_params; /* only important for PDDL where
				:VARS may be added to the param list
				which must be hidden when writing the
				plan to an output file */
  
  struct OP *next;
} *op_list, op_list_elt;


/* completely instantiated operator, corresponds to actions in the paper
 */
typedef struct OPERATOR {

  char *name;

  fact_list hooks;

  token_list objects;

  struct OPERATOR *next;
} *operator_list, operator_list_elt;


/* the type_tree structure is used to deal with types and subclasses
   of types */
typedef struct TYPETREE_LIST *type_tree_list, type_tree_list_elt;

typedef struct TYPETREE 
{
  token name;  /* an object type */
  type_tree_list sub_types;
} *type_tree, type_tree_elt;

struct TYPETREE_LIST 
{
  type_tree item;
  struct TYPETREE_LIST *next;
};

/*
 * function prototypes
 */

/* lex-fct.l */
char* rmdash( char* s );

void get_fct_file_name( char *filename );

/* scan-fct.tab.c */
void load_fct_file( char *filename );
void print_factlist( fact_list list );
void print_clauselist( fact_list list );
void print_tokenlist(token_list list);
void print_fct( fact_list con_list, token_list ini_list, token_list gol_list );

/* scan-ops.tab.c */
void load_ops_file( char *filename );
void print_ops( op_list op );

/* in file eq_preproc.c */
void preprocess_domain_vars();
void preprocess_goals();
BOOLEAN args_eq( token_list two_args );
token_list copy_tl_return_end( token_list tl, token_list *end );
void preprocess_pl1_facts( void );
op_list rec_prepoc_op_conditions( op_list op );
fact_list transform_to_atomic_negation(fact_list f,BOOLEAN neg );
fact_list eliminate_quantifiers( fact_list f );
fact_list rewrite_in_cnf( fact_list* cond );
fact_list get_variables( fact_list* last );
fact_list instantiate_and_copy_fl( token old, token_list new_list, 
			      fact_list source );
void replace_in_fl( token old, token new_tok, fact_list source );
fact_list replace_in_fact_list( token old, token new_tok, fact_list source,
				BOOLEAN shall_copy, fact_list *end );
op_list rec_get_normal_form_precond( op_list op );
op_list rec_remove_quants_from_precond( fact_list* cond,
					op_list orig_eff );
op_list transform_forall_precond( op_list op, fact_list params_list );
op_list multiply_preconds( op_list old1, op_list old2 );
void preprocess_axioms();
op_list conjuctive_goals( op_list the_ops, op_list *pred_address );

/* in file not_preproc.c */
token_list get_types_from_constlist(fact_list pred, fact_list var);
void append_to_effects(token name, token_list args, fact_list* effect);
void build_neg_predlist(fact_list pred, token_list f_types);
void store_fact_list(fact_list i, fact_list* start, fact_list* last);
void store_token_list(token_list i, token_list* start, token_list* last);
void transform_initials(void);
int not_in_list( token_list pred, fact_list list );
fact_list in_list( token_list pred, fact_list list );
void preprocess_not(void);
void build_type_list();
token_list find_consts_of_subtypes( type_tree tt );
token_list get_objects_of_type( token type_name );
fact_list rec_inst_pred( fact_list facts, token_list typelist );
fact_list instantiate_pred( token name, token_list typelist, char * init_token );

/* in file utilities.c */
void wrap_up();
void log_start(string name, int detail);
void log_end(string name);
int logged_time(string name);
void print_times();
BOOLEAN is_negated(fact_list fl);
BOOLEAN is_effect(fact_list fl);
BOOLEAN is_precond(fact_list fl);
fact_list find_token_list_in_clause(fact_list source, token_list target);
void my_free(void *ptr);
void my_free_all();
void * my_alloc(unsigned int nelem, unsigned int elsize);
token get_pred(token_list predicate);
void global_replace_cons_var(token old, token new_tok);
token_list find_token(token_list source, token target);
BOOLEAN is_constraint_form(fact_list fl);
BOOLEAN is_influence(fact_list fl);
BOOLEAN is_set(fact_list fl);
BOOLEAN is_primed_prune(token_list tl);
BOOLEAN token_lists_equal(token_list left, token_list right);
void negate_fact_list(fact_list fl);
fact_list new_fact_list(void);
token_list new_token_list(void);
void factlist_from_tokenlist( fact_list *fl, token_list tl );
token copy_token( token s );
token_list copy_token_list(token_list source);
int fact_list_length(fact_list source);
int clause_list_length(fact_list source);
fact_list copy_fact_list(fact_list source);
token_list copy_complete_token_list(token_list source, token_list *end);
fact_list copy_complete_fact_list(fact_list source, fact_list *end);
fact_list copy_complete_fact_clause( fact_list source, fact_list *end );
op_list copy_complete_op( op_list source );
void free_complete_token_list( token_list source );
void free_complete_fact_list( fact_list source );
void free_complete_op_list( op_list source );
token new_token(int size);
op_list new_axiom_op_list(void);
op_list new_op_list( token name );
operator_list new_operator_list( token name );
type_tree new_type_tree( token name );
type_tree_list new_type_tree_list( token name );
type_tree main_type_tree();
fact_list build_object_list_from_ttl( type_tree_list ttl, 
				      fact_list types_done );
token_list type_already_known( token name, fact_list types );
void build_orig_constant_list();
void add_to_type_tree( fact_list types, type_tree tree );
void print_type_tree( type_tree root, int indent );
type_tree find_branch( token name, type_tree root );
char* strupcase( char* from );
fact_list make_adl_fact( int c );
int get_adl_token( fact_list f );

/* in file util_inst.c */
void eliminate_timeless(fact_list * clause);
void rename_fluent_params( op_list op, int id );
void rename_influence_objects( fact_list hooks, int & id );
token token_from_token_list( token_list tlist );
token_list token_list_from_fact_list( fact_list flist );
fact_list fact_list_from_cnf_form( fact_list cnf_clauses );
token_list copy_replace_tlist( token_list tlist, char *dest, char *source );
fact_list copy_replace_flist( fact_list flist, char *dest, char *source );
op_list_elt copy_replace_op( op_list_elt op, char *dest, char *source );
op_list_elt copy_op( op_list_elt op );
fact_list copy_remove( fact_list flist, fact_list rm );
token make_name( op_list_elt op );
token_list make_objects( op_list_elt op );
void check_instantiation(fact_list f_list, char* action);

/* in file print_solution.cc */
void pretty_print(bool solved, hash_map<string, bool> & soln,
			   hash_map<string, double> & vsoln);

/* in file instantiate.c */
void instantiate_list( op_list ops, fact_list constants );
void instantiate( op_list_elt ops, fact_list constants );
void instantiate_first_parameter( op_list_elt op,
                                  fact_list constants );
void print_operators( operator_list operators );


/* in file ipp_to_cnf.c */
void build_cnf_stmt();

/*
 * global variables
 */


/* technical variables */
extern hash_map<string, Stats> stats;
extern char *act_filename;
extern int lineno;
extern int act_err;
extern char *act_err_par;
extern int bracket_count;

/* globally important variables */
extern int detail_level;
extern int compact_level;
extern bool bParallel;
extern bool bInfluences;
extern bool bMutexInfl;

/* domain constants */
extern char *gdomain_name;
extern int gproblem_length;
extern int gproblem_length_serial;
extern int gproblem_length_parallel;
extern op_list loaded_ops;/* loaded, uninstantiated operators */
extern operator_list operators;/* list of instantiated operators */
extern token_list initial_facts;/* the initial facts */
extern fact_list raw_goal_facts;/* not yet preprocessed goal facts */
extern fact_list domain_vars_list;/* Wolf - to store all variables */
extern fact_list predicates_list;/* Wolf - all predicates with types */
extern fact_list orig_constant_list;/* to store all typed objects */
extern fact_list orig_initial_facts;/* stores initials as fact_list */
extern fact_list global_object_fl;/* fast finding of types and their objs */
extern type_tree_list global_type_tree_list;/* type hierarchy (PDDL) */

class CNF_STMT;
extern CNF_STMT solver_stmt;

extern char problemName[MAX_LENGTH];

/* optimization */

#endif __IPP_H

