#include "stats.h"
#include <math.h>

Stats::Stats() : _count(0), _total(0), _squaredTotal(0), _max(0), _min(0)
{
  _print = COUNT | MEAN | STD;
}

void Stats::log(double data)
{
  // max
  if (_count == 0 || data > _max)
    _max = data;

  // min
  if (_count == 0 || data < _min)
    _min = data;

  // count
  _count++;

  // total
  _total += data;

  // squared total
  _squaredTotal += data * data;
}

double Stats::mean()
{
  if (_count == 0)
    return 0;

  return _total / _count;
}

double Stats::std()
{
  if (_count == 0)
    return 0;

  return sqrt((_squaredTotal - _total * _total / _count) / _count);
}

double Stats::max()
{
  return _max;
}

double Stats::min()
{
  return _min;
}

double Stats::count()
{
  return _count;
}

double Stats::total()
{
  return _total;
}

void Stats::clear()
{
  _count = 0;
  _total = 0;
  _squaredTotal = 0;
  _max = 0;
  _min = 0;
}

ostream & operator << (ostream & out, Stats & stats)
{
  if (stats._print & Stats::TOTAL)
    out << "Total: " << stats.total() << '\t';

  if (stats._print & Stats::COUNT)
    out << "Count: " << stats.count() << '\t';

  if (stats._print & Stats::MEAN)
    out << "Mean: " << stats.mean() << '\t';

  if (stats._print & Stats::STD)
    out << "Std: " << stats.std() << '\t';
 
  if (stats._print & Stats::MIN)
    out << "Min: " << stats.min() << '\t';
  
  if (stats._print & Stats::MAX)
    out << "Max: " << stats.max() << '\t';

  return out;
}

void Stats::setPrint(Stats::Printables print)
{
  _print = print;
}
