/* (C) Copyright 1997 Albert Ludwigs University Freiburg
 *     Institute of Computer Science
 *
 * All rights reserved. Use of this software is permitted for 
 * non-commercial research purposes, and it may be copied only 
 * for that use.  All copies must include this copyright message.
 * This software is made available AS IS, and neither the authors
 * nor the  Albert Ludwigs University Freiburg make any warranty
 * about the software or its performance. 
 */

#ifndef lint
static char rcsid[] = "$Id: util_inst.c,v 1.12 1998/11/30 23:14:32 wolf Exp $";
#endif /* lint */

/*
 * some helpfunctions for instantiating
 */

#include"ipp.h"/* defines, data structures, fn prototypes, global variables */

#include <assert.h>

#define RENAME_DIGITS 9 /* maximum number of digits to use in renaming
					  constraint variables */

void rename_fluent_params( op_list op, int id )
{
  /* note that all the labels for a given operator instance have
	the same number! Therefore, I can actually do this a bit more
	efficiently by going through once and renaming everything in
	the fluent list with the new number. Leave that to later. !!! */
  token_list dummy;
  token_list curr_param = copy_complete_token_list(op->fluent_params, &dummy);
  token t;
  
  /* !!! Wolf - bit of a memory leak (because of copy_replace_op) !!! */
  op->fluent_params = curr_param;
  while (curr_param) {
    t = new_token(strlen(curr_param->item) + RENAME_DIGITS + 1);
    sprintf(t, "%s%d%s", "?", id, curr_param->item + 1);
    
    replace_in_fl(curr_param->item, t, op->hooks);
    my_free(curr_param->item);
    curr_param->item = t;

    curr_param = curr_param->next;
  }
}

void rename_influence_objects( fact_list hooks, int & id )
{
  token t;
  
  while (hooks) {
    if (is_influence(hooks)) {
	 /* replace the old object of the influence with a unique 
	    numbered object */
	 t = new_token(strlen(hooks->item->next->next->item) + RENAME_DIGITS + 2);
	 sprintf(t, "%s0%d%s", "?", id++, 
		    hooks->item->next->next->item + 1);
	 
	 my_free(hooks->item->next->next->item);
	 hooks->item->next->next->item = t;
    }

    hooks = hooks->next;
  }
}

token token_from_token_list( token_list tlist )
{
  int len = 0;
  token_list curr;
  token result = NULL;

  if (!tlist)
    return NULL;

  for (curr = tlist; curr; curr = curr->next) {
    len += strlen(curr->item) + 1;
  }

  result = new_token(len);
  strcpy(result, "");
  for (curr = tlist; curr; curr = curr->next) {
    strcat(result, curr->item);
    if (curr->next)
	 strcat(result, CONNECTOR);
  }

  return result;
}


/* converts list of ( list of strings ) into list of strings
 * by connecting items of ( list of strings ) with the CONNECTOR character)
 */
token_list token_list_from_fact_list( fact_list flist )
{
  token_list result = NULL, current, temp;
  char help[MAX_LENGTH] = "";
  
  if ( !flist ) return NULL;

  while (flist) {
    for ( current = flist->item; current; current = current->next ) {
	 strcat( help, current->item );
	 if ( current->next ) 
	   strcat( help, CONNECTOR );
    }
    
    temp = new_token_list();

    temp->item = new_token(strlen(help) + 1);
    strcpy( temp->item, help );
    
    temp->next = result;
    result = temp;
  }

  return result;
}

fact_list fact_list_from_cnf_form(fact_list cnf_clauses)
{
  fact_list result = NULL, temp;
  token_list current = NULL, tl;

  while (cnf_clauses) {
    assert(get_adl_token(cnf_clauses) == OR_CONST);
    cnf_clauses = cnf_clauses->next;
    
    current = NULL;

    /* collect the tokens of this token_list/clause */
    while (get_adl_token(cnf_clauses) != RBRACK_CONST) {
	 temp = cnf_clauses;
	 cnf_clauses = cnf_clauses->next;
	 assert(cnf_clauses);

	 tl = new_token_list();
	 tl->item = token_from_token_list(temp->item);
	 tl->next = current;
	 current = tl;
    }

    temp = new_fact_list();
    temp->item = current;
    temp->next = result;
    result = temp;

    cnf_clauses = cnf_clauses->next;
  }

  return result;
}

/* second inner helpfunction for copy_replace,
 * dealing with a fact_list
 */
fact_list copy_replace_flist( fact_list flist, char *dest, char *source )

{
  fact_list result, dummy;

  result = copy_complete_fact_list(flist, &dummy);
  replace_in_fl(dest, source, result);

  return result;
}

/* makes copy of given uninstantiated operator op,
 * replacing string dest with string source whenever
 * finding the first; removing first parameter from list;
 * thereby instantiates op's first parameter with source
 * ( cause dest = first parameter, see instantiate_first_parameter 
 * in instantiate.c )
 */
op_list_elt copy_replace_op( op_list_elt op, char *dest, char *source )

{

  op_list_elt result;
  fact_list temp1;
  token_list temp2, temp3;

  temp1 = new_fact_list();
  temp2 = new_token_list();
  temp3 = new_token_list();
  temp3->item = source;
  temp3->next = NULL;
  temp2->item = dest;
  temp2->next = temp3;
  temp1->item = temp2;
  temp1->next = op.params_objects;
  result.params_objects = temp1;

  result.fluent_params = op.fluent_params;

  result.params = op.params;

  result.name = op.name;

  result.hooks = copy_replace_flist( op.hooks, dest, source );

  result.next = NULL;

  return result;

}


op_list_elt copy_op( op_list_elt op )

{

  op_list_elt result;

  result.params = op.params;

  result.fluent_params = op.fluent_params;

  result.name = op.name;

  result.hooks = op.hooks;

  result.params_objects = op.params_objects;

  result.next = NULL;

  return result;

}


/* makes copy of factlist, removing given item of it
 *
 * is used by instantiate_first_parameter ( in instantiate.c )
 * and build_inst_effect_list( same file )
 */
fact_list copy_remove( fact_list flist, fact_list rm )

{

  fact_list temp;

  if ( !flist ) return NULL;

  temp = new_fact_list();
  if ( flist == rm ) {
    if ( flist->next ) { 
      temp->item = flist->next->item;
      temp->next = flist->next->next;
    } else {
      my_free( temp );
      temp = NULL;
    }  
  } else {
    temp->item = flist->item;
    temp->next = copy_remove( flist->next, rm );
  }

  return temp;

}

token make_name( op_list_elt op )
{

  token result;
  int len = 0;
  fact_list i_param, i_fact;
  token_list i_fparam;

  len = strlen(ACTION_LABEL) + strlen(op.name) + 1;
  for ( i_param = op.params; i_param; i_param = i_param->next ) {
    for ( i_fact = op.params_objects; i_fact; i_fact = i_fact->next ) {
	 if ( strcmp( i_fact->item->item, i_param->item->item ) == SAME )
	   break;
    }
    assert( i_fact != NULL );
    len += strlen(i_fact->item->next->item) + 1;
  }

  for ( i_fparam = op.fluent_params; i_fparam; i_fparam = i_fparam->next ) {
    len += strlen( i_fparam->item ) + 1;
  }

  result = new_token(len);
  strcpy(result, ACTION_LABEL);
  strcat(result, op.name);
  for ( i_param = op.params; i_param; i_param = i_param->next ) {
    for ( i_fact = op.params_objects; i_fact; i_fact = i_fact->next ) {
	 if ( strcmp( i_fact->item->item, i_param->item->item ) == SAME )
	   break;
    }
    assert( i_fact != NULL );
    strcat( result, CONNECTOR );
    strcat( result, i_fact->item->next->item );
  }

  for ( i_fparam = op.fluent_params; i_fparam; i_fparam = i_fparam->next ) {
    strcat( result, CONNECTOR );
    strcat( result, i_fparam->item );
  }

  return result;
}


token_list make_objects( op_list_elt op )

{

  token_list result = NULL, temp;
  fact_list i_ob;

  for ( i_ob=op.params_objects; i_ob; i_ob=i_ob->next ) {
    temp = new_token_list();
    temp->item = i_ob->item->next->item;
    temp->next = result;
    result = temp;
  }

  return result;

}

void check_instantiation(fact_list f_list, char* action)
{
  /* !!! need to search for uninstantiated constraint parameters !!! */
  token_list t_token;
  fact_list t_fact;

  for (t_fact = f_list; t_fact != NULL; t_fact = t_fact->next)
    {
	 /* constraints may have variables still */
	 if (is_constraint_form(t_fact))
	   continue;
      for (t_token = t_fact->item; t_token != NULL; t_token = t_token->next)
	   {
		if ('?' == *(t_token->item))
		  {
		    fprintf(stderr, "\nipp: uninstantiated variable found at:");
		    if (action)
			 fprintf(stderr, "\nipp: operator \"%s\" > %s", action, t_token->item);
		    else
			 fprintf(stderr, "\nipp: all-quantified goal in fact file > %s", t_token->item);
		    fprintf(stderr, "\nipp: Check your input files for typing errors.\n\n");
		    OUTPUT_FILE;
		    exit(1);
		  }
	   }
    }
}

/* Really only eliminates eq/!eq... should eliminate timeless!!! */
void eliminate_timeless(fact_list * clause)
{
    fact_list last_clause = NULL;
    fact_list current = *clause;
    fact_list this_clause = *clause;
    BOOLEAN true_clause = FALSE;

    assert (clause);

    /* loop over the OR clauses until reach the end */
    while (this_clause) {
	   current = this_clause;
	   assert(get_adl_token(current) == OR_CONST);
	   true_clause = FALSE;

	   /* *clause is beginning of stmt, this_clause is beginning of
		 this OR clause, current == this_clause, last_clause is end of
		 last clause (RBRACK_CONST) */
	   while (get_adl_token(current->next) != RBRACK_CONST) {
		  token tok = get_pred(current->next->item);

		  if (strcmp(tok, EQ_STR) == SAME) {
			 BOOLEAN pos = current->next->item->item[0] != '!';
			 token_list args = 
			   (!pos ? 
			    (strcmp(current->next->item->item + 1, EQ_STR) == SAME) :
			    (strcmp(current->next->item->item, EQ_STR) == SAME)) ?
				 current->next->item->next :
			 current->next->item->next->next;
			 BOOLEAN eq_args = args_eq(args);

			 /* case where this predicate is true */
			 if (pos == eq_args) {
				fact_list temp = this_clause;

				true_clause = TRUE;

				while (get_adl_token(current) != RBRACK_CONST) {
				    current = current->next;
				}
				this_clause = current->next;
				current->next = NULL;
				free_complete_fact_list(temp);
				if (last_clause)
				  last_clause->next = this_clause;
				else
				  *clause = this_clause;
				break;
			 }
			 /* case where this predicate is false */
			 else {
				fact_list temp = current->next;
				current->next = current->next->next;
				temp->next = NULL;
				free_complete_fact_list(temp);
			 }
		  }
		  else
		    current = current->next;
	   }
	   /* empty clause eliminates whole shebang! */
	   if (!true_clause && get_adl_token(this_clause->next) == RBRACK_CONST) {
		  free_complete_fact_list(*clause);
		  *clause = NULL;
		  return;
	   }

	   /* set up the last_clause/this_clause appropriately */
	   if (!true_clause) {
		  last_clause = current->next;
		  this_clause = current->next->next;
	   }
    }
}
