/* (C) Copyright 1997 Albert Ludwigs University Freiburg
 *     Institute of Computer Science
 *
 * All rights reserved. Use of this software is permitted for 
 * non-commercial research purposes, and it may be copied only 
 * for that use.  All copies must include this copyright message.
 * This software is made available AS IS, and neither the authors
 * nor the  Albert Ludwigs University Freiburg make any warranty
 * about the software or its performance. 
 */

/* 	$Id: utilities.c,v 1.15 1999/10/19 06:20:02 wolf Exp wolf $	 */

#ifndef lint
static char vcid[] = "$Id: utilities.c,v 1.15 1999/10/19 06:20:02 wolf Exp wolf $";
#endif /* lint */


#include"ipp.h"
#include<ctype.h>
#include<assert.h>
#include<hash_set>

/* These are helper functions that might be used everywhere */

/* Functions dealing with time logging */

typedef hash_map<string, pair<bool, unsigned int> > TIME_LOG;

TIME_LOG time_log;
list<string> time_names;

/* call this to access time logged under "name" */
int logged_time(string name)
{
  if (time_log.find(name) == time_log.end())
    return 0;
  else
    return (int)time_log[name].second;
}

/* call this to start (or re-start) the time on "name" */
void log_start(string name, int detail)
{
  if (detail_level >= detail) {
    struct tms tms_time;
    int time;
    

    if (time_log.find(name) == time_log.end()) {
	 times(&tms_time);
    
	 time = tms_time.tms_utime + tms_time.tms_stime;

	 time_log[name] = pair<bool, unsigned int>(false, time);
	 time_names.push_back(name);
    }
    else if (time_log[name].first == true) {
	 /* push the start time into the past to add the amount of time
	    already registered in */
	 times(&tms_time);
	 
	 time = tms_time.tms_utime + tms_time.tms_stime;

	 time_log[name] = 
	   pair<bool, unsigned int>(false, time - time_log[name].second);
    }
  }
}

/* call this once to end the time on "name" */
void log_end(string name)
{
  /* if the entry isn't in there, assume it's because the detail
	level wasn't high enough to merit storing it */

  struct tms tms_time;
  int time;
  
  times(&tms_time);
  
  time = tms_time.tms_utime + tms_time.tms_stime;

  if (time_log.find(name) != time_log.end()) {
    /* make sure this is the first time this has been logged over */
    assert(time_log[name].first == false);

    time_log[name] = pair<bool, unsigned int>(true, 
									 time - time_log[name].second);
  }
}

/* call this to print out the times */
void print_times()
{
  if (detail_level == DETAIL_NONE)
    return;

  list<string>::iterator i;

  cout << "Timing information" << endl;

  for (i = time_names.begin(); i != time_names.end(); i++) {
    string name = (*i);
    if (time_log.find(name) == time_log.end())
	 continue;

    bool closed = time_log[name].first;
    unsigned int time = time_log[name].second;

    if (!closed)
	 continue;

    cout << name << ":\t" << time / 100.0 << " s" << endl;
  }
}

hash_set<unsigned long> mem_set;
unsigned long mem_used = 0;

/* my_alloc replaces calloc and allows _all_ memory to be freed at once */
/* my_free replaces free and refs the structure updated by my_alloc */
/* my_free_all frees all memory recorded in my_alloc */
void my_free(void *ptr)
{
  /*  if (mem_set.erase((unsigned long)ptr) > 0) {
    free(ptr);
  }*/
  free (ptr);
}

void my_free_all()
{
  /*
  hash_set<unsigned long>::iterator i;
  for (i = mem_set.begin(); i != mem_set.end(); i++) {
    free((void*)(*i));
  }

  mem_set.clear();
  */
}

void * my_alloc(unsigned int nelem, unsigned int elsize)
{
  /*
  mem_used += nelem * elsize;
  */
  void * result = calloc(nelem, elsize);
  CHECK_MEMORY(result);
  /*
  mem_set.insert((unsigned long) result);
  */
  return result;
}

/* returns the predicate in a token_list, skipping EFFECT/PRECOND and ! */
/* assumes this is a literal */
token get_pred(token_list predicate)
{
    BOOLEAN negated = predicate->item[0] == '!';

    assert(predicate && predicate->item);

    if (negated ? 
	   (strcmp(predicate->item + 1, EFFECT_STR) == SAME ||
	    strcmp(predicate->item + 1, PRECOND_STR) == SAME) :
	   (strcmp(predicate->item, EFFECT_STR) == SAME ||
	    strcmp(predicate->item, PRECOND_STR) == SAME)) {
	   assert(predicate->next);
	   return predicate->next->item;
    }
    else {
	   return negated ? predicate->item + 1 : predicate->item;
    }
}

/* search for positive SETs/INFLUENCEs whose first arg is t */
BOOLEAN change_in_fl(token t, fact_list fl)
{
  while (fl) {
    if (!fl->item || !fl->item->item || !fl->item->next ||
	   !fl->item->next->item || !fl->item->next->next || 
	   !fl->item->next->next->item) {
	 fl = fl->next;
	 continue;
    }

    if (strcmp(fl->item->item, EFFECT_STR) == SAME &&
	   strcmp(fl->item->next->item, SET_STR) == SAME &&
	   strcmp(fl->item->next->next->item, t) == SAME)
	 return TRUE;

    /* compare against the un-renumbered var on the right hand
	  side of the INFLUENCE */
    if (strcmp(fl->item->item, EFFECT_STR) == SAME &&
	   strcmp(fl->item->next->item, INFLUENCE_STR) == SAME &&
	   strcmp(fl->item->next->next->next->next->item, t) == SAME)
	 return TRUE;
    
    fl = fl->next;
  }

  return FALSE;
}

void global_replace_cons_var(token old, token new_tok)
{
  operator_list i_op;

  if (change_in_fl(old, orig_initial_facts)) {
    fprintf(stderr, "ERROR: %s is a constant, but is set in initials.\n",
		 old);
    OUTPUT_FILE;
  }

  replace_in_fl(old, new_tok, orig_initial_facts);

  /* goals are PRECONDS, so there can't be any SET stmts */
  replace_in_fl(old, new_tok, raw_goal_facts);

  for (i_op = operators; i_op; i_op = i_op->next) {
    /* !!! scoping is screwed up here, global can overwrite
	  local, not a really big deal, though !!! */
    /* first check for sets (error if there are any), 
	  then replace all occurences of old in the hooks */
    if (change_in_fl(old, i_op->hooks)) {
	 fprintf(stderr, "ERROR: %s is declared constant, but set in op: %s.\n",
		   old, i_op->name);
	 OUTPUT_FILE;
    }

    replace_in_fl(old, new_tok, i_op->hooks);
  }
}

fact_list new_fact_list(void)
{
  fact_list result;

  result = ( fact_list ) my_alloc( 1, sizeof( fact_list_elt ) );
  CHECK_MEMORY(result);
/*   if (!result) */
/*     { */
/*       fprintf(stdout, "%s", NO_MEMORY); */
/*       fprintf(outputFile, "%s", NO_SOLUTION); */
/*       exit(2); */
/*     } */

  result->item = NULL; 
  result->next = NULL;
  return result;
}

token_list new_token_list(void)
{
  token_list result;
  result = ( token_list ) my_alloc( 1, sizeof( token_list_elt ) );
  CHECK_MEMORY(result);
/*   if (!result) */
/*     { */
/*       fprintf(stdout, "%s", NO_MEMORY);  */
/*       fprintf(outputFile, "%s", NO_SOLUTION); */
/*       exit(2); */
/*     } */

  result->item = NULL; 
  result->next = NULL;
  return result;
}

token new_token(int size)
{
  char* tok;

  tok = (char*) my_alloc(size, sizeof(char));
  CHECK_MEMORY(tok);
/*   if (!tok) */
/*     { */
/*       fprintf(stdout, "%s", NO_MEMORY); */
/*       fprintf(outputFile, "%s", NO_SOLUTION); */
/*       exit(2); */
/*     } */
  return tok;
}

BOOLEAN is_negated(fact_list fl)
{
  if (!fl || !fl->item || !fl->item->item)
    return FALSE;

  return fl->item->item[0] == '!';
}

BOOLEAN is_first_token(fact_list fl, token t)
{
  if (!fl || !fl->item || !fl->item->item)
    return FALSE;

  return strcmp(fl->item->item, t) == SAME;
}

BOOLEAN is_effect(fact_list fl)
{
  return is_first_token(fl, EFFECT_STR) ||
    is_first_token(fl, NOT_EFFECT_STR);
}

BOOLEAN is_precond(fact_list fl)
{
  return is_first_token(fl, PRECOND_STR) ||
    is_first_token(fl, NOT_PRECOND_STR);
}

/* Note that this will return true iff fl is a _positive_ constraint */
BOOLEAN is_constraint_form(fact_list fl)
{
  if (!fl || !fl->item || !fl->item->next || 
	 !fl->item->next->item)
    return FALSE;

  return ((strcmp(fl->item->next->item, TEST_STR) == SAME) ||
		(strcmp(fl->item->next->item, SET_STR) == SAME) ||
		(strcmp(fl->item->next->item, INFLUENCE_STR) == SAME));
}

BOOLEAN is_influence(fact_list fl)
{
  if (!fl || !fl->item || !fl->item->next || 
	 !fl->item->next->item)
    return FALSE;

  return (strcmp(fl->item->next->item, INFLUENCE_STR) == SAME);
}

BOOLEAN is_set(fact_list fl)
{
  if (!fl || !fl->item || !fl->item->next || 
	 !fl->item->next->item)
    return FALSE;

  return (strcmp(fl->item->next->item, SET_STR) == SAME);
}

BOOLEAN is_primed_prune(token_list tl)
{
  token t;

  while (tl) {
    t = tl->item;
    if (!t)
	 continue;
    while (*t) {
	 if (*t == '\'')
	   break;
	 t++;
    }
    if (*t == '\'') {
	 token rem = new_token(strlen(tl->item));
	 *t = '\0';
	 strcpy(rem, tl->item);
	 strcat(rem, t + 1);
	 my_free(tl->item);
	 tl->item = rem;

	 return true;
    }
    tl = tl->next;
  }

  return false;
}

fact_list find_token_list_in_clause(fact_list source, token_list target)
{
  fact_list fl;

  for (fl = source->next; get_adl_token(fl) != RBRACK_CONST; fl = fl->next) {
    if (token_lists_equal(fl->item, target))
	 return fl;
  }

  return NULL;
}

token_list find_token(token_list source, token target)
{
  token_list tl;


  for (tl = source; tl; tl = tl->next) {
    if (strcmp(tl->item, target) == SAME) {
	 /* found the target, break out of the loop */
	 break;
    }
  }
  
  /* tl is the target, or, if the target is not in source, tl is NULL */
  return tl;
}

/* check two token_lists for surface equality (i.e., not necessarily the same
 * memory, but the same values)
 */
BOOLEAN token_lists_equal(token_list left, token_list right)
{
  for ( ; left && right; left = left->next, right = right->next) {
    if (left->item == NULL || right->item == NULL) {
	 if (left->item == NULL && right->item == NULL)
	   continue;
	 else
	   return FALSE;
    }
    else {
	 if (strcmp(left->item, right->item) != SAME) {
	   return FALSE;
	 }
    }
  }

  return left == NULL && right == NULL;
}

/* return the token matching the target in source, removing it from source,
 * or return NULL if no match is found. */
token_list remove_token(token_list * source, token target)
{
  token_list tl;

  if (source == NULL)
    return NULL;

  /* handle the special case where the target is the first token */
  if (strcmp((*source)->item, target) == SAME) {
    tl = *source;
    *source = (*source)->next;
    tl->next = NULL;
    return tl;
  }

  /* otherwise, the target is later in the list */
  for (tl = *source; tl->next; tl = tl->next) {
    if (strcmp(tl->next->item, target) == SAME) {
	 /* found the target, break out of the loop */
	 break;
    }
  }
  
  /* tl->next is the token sought or, if that token isn't in source,
   * tl->next is NULL. */

  if (tl->next) {
    token_list result = tl->next;
    tl->next = result->next;
    result->next = NULL;
    return result;
  }
  else {
    return NULL;
  }
}

op_list new_op_list( token name )
{
  op_list act_op;
  act_op = ( op_list ) my_alloc( 1, sizeof( op_list_elt ) );
  CHECK_MEMORY(act_op);
  if ( name )
    {
      act_op->name = new_token( strlen( name ) + 1 );
      strcpy( act_op->name, name );
    }
  else
    name = NULL;
  act_op->hooks = NULL;
  act_op->params = NULL;
  act_op->fluent_params = NULL;
  act_op->params_objects = NULL;
  act_op->next = NULL;
  act_op->number_of_real_params = 0; /* only for PDDL to hide some params */
  return act_op;
}

operator_list new_operator_list( token name )
{
  operator_list act_op;
  act_op = ( operator_list ) my_alloc( 1, sizeof( operator_list_elt ) );
  CHECK_MEMORY(act_op);
  act_op->name = new_token( strlen( name ) + 1 );
  strcpy( act_op->name, name );
  act_op->hooks = NULL;
  act_op->objects = NULL;
  act_op->next = NULL;
  return act_op;
}

type_tree new_type_tree( token name )
{
  type_tree act_type;
  
  if (!name)
    return NULL;
  act_type = ( type_tree ) my_alloc( 1, sizeof( type_tree_elt ) );
  CHECK_MEMORY(act_type);
  act_type->name = new_token( strlen( name ) + 1 );
  strcpy( act_type->name, name );
  act_type->sub_types = NULL;
  return act_type;
}

type_tree_list new_type_tree_list( token name )
{
  type_tree_list act_type_list;
  
  act_type_list = ( type_tree_list ) my_alloc( 1, sizeof( type_tree_list_elt ) );
  CHECK_MEMORY(act_type_list);
  if ( name )
    act_type_list->item = new_type_tree( name );
  else
    act_type_list->item = NULL;
  act_type_list->next = NULL;
  
  return act_type_list;
}

type_tree main_type_tree()
{
  type_tree_list ttl;

  for ( ttl = global_type_tree_list; ttl; ttl = ttl->next )
    if ( strcmp( ttl->item->name, STANDARD_TYPE ) == SAME )
      return ttl->item;
  return NULL;
}

/* steps recursively through type tree and searches for name */
type_tree find_branch( token name, type_tree root )
{
  type_tree p;
  type_tree_list ttl;

  if ( !root )
    return NULL;
  if ( strcmp( root->name, name ) == SAME )
    return root;
  if ( !root->sub_types )
    return NULL;
  for ( ttl=root->sub_types; ttl; ttl=ttl->next )
    if ( (p = find_branch( name, ttl->item )) )
      return p;
  return NULL; 
}

/* calls itself recursively to get all objects that are of the types and
   subtypes of ttl */
fact_list build_object_list_from_ttl( type_tree_list ttl, 
				      fact_list types_done )
{
  fact_list f, td, std;
  token_list t, tl_dummy, st;
  type_tree_list sttl;

  if ( !ttl )
    return types_done;
  
  types_done = build_object_list_from_ttl( ttl->next, types_done );

  if ( (t = type_already_known( ttl->item->name, types_done )) )
    return types_done;

  td = new_fact_list();
  t = td->item = new_token_list();
  /* begin with the name of the type... */
  t->item = copy_token( ttl->item->name );
  for ( f = orig_constant_list; f; f = f->next )
    {
      /* ...followed by objects of that type. */
      if ( strcmp( f->item->next->item, ttl->item->name ) == SAME )
	{
	  t->next = new_token_list();
	  t = t->next;
	  t->item = copy_token( f->item->item );
	}
    }
  /* now append the objects of the subtypes */
  std =  build_object_list_from_ttl( ttl->item->sub_types, types_done );
  /* now we can be sure that for each subtype a list with all
     objects of that type is somewhere in std. We simply take these
     lists and copy all of them into a new one for the supertype */
  for ( sttl = ttl->item->sub_types; sttl; sttl = sttl->next )
    {
      st = type_already_known( sttl->item->name, std );
      if ( st )
	{
	  t->next = copy_complete_token_list( st, &tl_dummy );
	  t = tl_dummy;
	}
    }
  td->next = std;
  return td;
}

token_list type_already_known( token name, fact_list types )
{
  fact_list f;

  for ( f = types; f; f = f->next )
    if ( strcmp( f->item->item, name ) == SAME )
      return f->item->next;
  return NULL;
}


void build_orig_constant_list()
{
  fact_list f, nextf, end = NULL;
  token_list t;
  BOOLEAN do_count;

  global_object_fl = build_object_list_from_ttl( global_type_tree_list, 
									    NULL );
  free_complete_fact_list( orig_constant_list );
  orig_constant_list = NULL;
  for ( f = global_object_fl; f; f = nextf )
    {
      nextf = f->next;
      if ( strcmp( f->item->item, STANDARD_TYPE ) == SAME )
	   do_count = TRUE;
      else
	   do_count = FALSE;
      for ( t = f->item->next; t; t = t->next )
	   {
		if ( !orig_constant_list )
		  orig_constant_list = end = new_fact_list();
		else
		  {
		    end->next = new_fact_list();
		    end = end->next;
		  }
		end->item = new_token_list();
		end->item->item = copy_token( t->item );
		end->item->next = new_token_list();
		end->item->next->item = copy_token( f->item->item );
	   }
    }
}


void add_to_type_tree( fact_list t_list, type_tree tree )
{
  type_tree branch = tree;
  type_tree_list new_branch;
  token this_type; 
  token super_type;

  /* step through list and build a hierarchy of types */
  for( ; t_list; t_list=t_list->next )
    {
      this_type = t_list->item->item;
      if ( !t_list->item->next ) 
	{
	  fprintf(stderr, "\n%s: error at '%s'.\n", act_filename, this_type );
	  OUTPUT_FILE;
	  exit( 1 );
	}
      super_type = t_list->item->next->item;
      if ( strcmp( branch->name, super_type ) != SAME )
	branch = find_branch( super_type, tree );
      if ( !branch ) 
	{
	  fprintf(stderr, "\n%s: unknown type '%s'.\n", act_filename, 
			super_type );
	  OUTPUT_FILE;
	  exit( 1 );
	}
      /* now the type is a subtype of the one currently looked at 
	 in the type tree */
      new_branch = new_type_tree_list( this_type );
      new_branch->next = branch->sub_types;
      branch->sub_types = new_branch;
    }
}


void print_type_tree_list( type_tree_list root, int indent )
{
  int i;
  for(i=0;i<2*indent; i++ ) printf( " " );
  if ( root ) 
    {
      if ( !root->item || !root->item->name )
	{
	  fprintf(stderr, "\n%s: internal error: no type name specified.\n", 
		  act_filename );
	  OUTPUT_FILE;
	  exit( 1 );
	}
      else
	printf( "%s\n", root->item->name );
      if ( root->item->sub_types )
	print_type_tree_list( root->item->sub_types, indent+1 );
      if ( root->next )
	print_type_tree_list( root->next, indent );
    }
}

/* 'blowing up' a token list to get a fact list in which all
   token_lists only contain one token ... 
   this is only needed to be able to pass operator instatiation
   even when a (ex-goal) operator in 'uninstatiated format doesn't 
   need instatiation at all */
void factlist_from_tokenlist( fact_list *fl, token_list tl )
{
  fact_list *act_f;
  token_list act_t;

  for( act_f=fl, act_t=tl; act_t; 
       act_f = &((*act_f)->next), act_t=act_t->next )
    {
      *act_f = new_fact_list();
      (*act_f)->item = new_token_list();
      (*act_f)->item->item = act_t->item;
    }
}

token copy_token( token s )
{
  token d;
  
  d = new_token( strlen( s ) + 1 );
  strcpy( d, s );
  return d;
}

/*
 * This function copies a token_list by creating a new
 * token_list_elt for every element in source and
 * copying the item and next pointer to the new token_list_elt.
 * Then the function calls itself with the next element of source.
 *
 * returns: the start of the new token_list
 */
token_list copy_token_list( token_list source )
{
  token_list dummy;

  return copy_complete_token_list( source, &dummy );
  /* if it were sure that copy_token_list() would be used only
     when the tokens are never changed, copy_token_list() could
     be implemented without strcpy() */
}

token_list copy_complete_token_list( token_list source, token_list *end )
{
  token_list result = NULL, temp;

  if (!source) {
    *end = NULL;
    return NULL;
  }

  while (source) {
    temp = new_token_list();
    if (source->item) {
	 temp->item = new_token(strlen(source->item) + 1);
	 strcpy(temp->item, source->item);
    }

    if (!result) {
	 result = *end = temp;
    }
    else {
	 (*end)->next = temp;
	 *end = temp;
    }
    
    source = source->next;
  }

  return result;
}

int fact_list_length( fact_list source )
{
  int len = 0;
  while (source) {
    len++;
    source = source->next;
  }

  return len;
}

int clause_list_length( fact_list source )
{
  int len = 0;
  while (source) {
    if (get_adl_token(source) == LIT_CONST)
	 len++;
    source = source->next;
  }

  return len;
}

fact_list copy_fact_list( fact_list source )
{
  fact_list result = NULL, temp, end = NULL;

  if (!source) {
    return NULL;
  }

  while (source) {
    temp = new_fact_list();
    temp->item = source->item;

    if (!result) {
	 result = end = temp;
    }
    else {
	 end->next = temp;
	 end = temp;
    }

    source = source->next;
  }

  return result;
}

/* negates all LIT_CONSTs */
void negate_fact_list(fact_list fl)
{
  token temp;
  for ( ; fl; fl = fl->next) {
    if (get_adl_token(fl) != LIT_CONST)
	 continue;

    if (fl->item->item[0] == '!') {
	 temp = new_token(strlen(fl->item->item));
	 strcpy(temp, fl->item->item + 1);
	 my_free(fl->item->item);
	 fl->item->item = temp;
    }
    else {
	 temp = new_token(strlen(fl->item->item) + 2);
	 sprintf(temp, "%s%s", "!", fl->item->item);
	 my_free(fl->item->item);
	 fl->item->item = temp;
    }
  }
}

fact_list copy_complete_fact_list( fact_list source, fact_list *end )
{
  fact_list result = NULL, temp;
  token_list dummy;

  if (!source) {
    *end = NULL;
    return NULL;
  }

  while (source) {
    temp = new_fact_list();
    temp->item = copy_complete_token_list(source->item, &dummy);

    if (!result) {
	 result = *end = temp;
    }
    else {
	 (*end)->next = temp;
	 *end = temp;
    }

    source = source->next;
  }

  return result;
}

/* copy only to a right bracket */
fact_list copy_complete_fact_clause( fact_list source, fact_list *end )
{
  fact_list result = NULL, temp;
  token_list dummy;

  if (!source || get_adl_token(source) == RBRACK_CONST) {
    *end = NULL;
    return NULL;
  }

  while (source && get_adl_token(source) != RBRACK_CONST) {
    temp = new_fact_list();
    temp->item = copy_complete_token_list(source->item, &dummy);

    if (!result) {
	 result = *end = temp;
    }
    else {
	 (*end)->next = temp;
	 *end = temp;
    }

    source = source->next;
  }

  return result;
}

op_list copy_complete_op( op_list source )
{
  op_list temp;
  fact_list dummy;
  token_list tl_dummy;

  if ( !source )
    {
      temp = NULL;
    }
  else
    {
      temp = new_op_list( source->name );
      temp->params = 
	   copy_complete_fact_list( source->params, &dummy );
	 temp->fluent_params = 
	   copy_complete_token_list( source->fluent_params, &tl_dummy );
      temp->hooks =
	   copy_complete_fact_list( source->hooks, &dummy );
      temp->number_of_real_params =
	   source->number_of_real_params;
      temp->next = NULL;
    }

  return temp;
}

void free_complete_token_list( token_list source )
{
  token_list tl;
  while (source) {
    tl = source;
    source = source->next;
    if (tl->item)
	 my_free(tl->item);
    my_free(tl);
  }
}

void free_complete_fact_list( fact_list source )
{
  fact_list fl;
  while ( source )
    {
	 fl = source;
	 source = source->next;
	 free_complete_token_list(fl->item);
	 my_free(fl);
    }
}

void free_complete_op_list( op_list source )
{
  /*
  if ( source )
    {
	 free_complete_op_list( source->next );
	 free_complete_fact_list( source->parameters );
	 free_complete_fact_list( source->hooks );
	 free_complete_token_list( source->fluent_params );
	 my_free( source );
    }
  return;
  */
}

char* strupcase( char* from )
{
  char* to;
  char* res = (char*) my_alloc( strlen( from )+1, sizeof(char) );
  CHECK_MEMORY(res);
  to = res;
  if ( !(*from) ) return NULL;
  for( ; *from; to++, from++ )
    *to = (char) toupper( (int) *from );
  *to = 0;
  return res;
}


fact_list make_adl_fact( int c )
{
  fact_list result;

  result = new_fact_list();
  result->item = new_token_list();
  result->item->item = new_token( 2 );
  result->item->item[0] = (char) c;
  result->item->item[1] = 0;
  return result;
}

/* Wolf - account for pre and effect prepends */
int get_adl_token( fact_list f )
{
  if ( ( !f ) || ( !f->item ) || ( !f->item->item ) )
    return 0;
  else if ( (int) f->item->item[0] > ENDNOT_CONST ) 
    {
	 token_list tl = f->item;

	 if ( tl->item[0] == '?' )
	   return VAR_CONST;
	 else
	   return LIT_CONST;
    }
  else
    return (int) f->item->item[0];
}

/* print out stats */
void print_stats()
{
  if (detail_level < DETAIL_STATS)
    return;

  hash_map<string, Stats>::iterator i;

  Stats::Printables full_stat, stat, count;

  count = Stats::COUNT;
  stat = (Stats::Printables)(count | Stats::MEAN | Stats::STD);
  full_stat = (Stats::Printables)(stat | Stats::MAX | Stats::MIN);

  stats["Successful constraint adds"].setPrint(stat);
  stats["Constraint adds"].setPrint(full_stat);
  stats["Constraint deletes"].setPrint(full_stat);
  stats["Restarts"].setPrint(count);
  stats["Labelled variables"].setPrint(full_stat);
  stats["Learn clause"].setPrint(full_stat);
  stats["Solver loop"].setPrint(count);
  stats["Branch points"].setPrint(stat);
  stats["Heuristic constraint"].setPrint(count);
  stats["Heuristic contradiction"].setPrint(count);
  stats["Constraint backup clause"].setPrint(full_stat);
  stats["Worthless count"].setPrint(count);

  cout << "Statistics:" << endl;
  for (i = stats.begin(); i != stats.end(); i++) {
    cout << '\t' << (*i).first << '\t' << (*i).second << endl;
  }
}

/* print out final info */
void wrap_up()
{
  print_times();
  print_stats();
}
