#include "Cl.h"
#include <iostream.h>
#include <fstream.h>
#include <strstream.h>
#include <set.h>

extern ClConstraint * parseConstraint(istream & in, ClVariable * aVars);

string itos(int i)
  // pre: i > 0
{
  string s = "";

  while (i)
    {
	 s = ((char) (i%10 + '0')) + s;
	 i /= 10;
    }

  return s;
}

ClVariable * createVars(int num)
{
  ClVariable * array = new ClVariable[num];
  for (int i = 0; i < num; i++) {
    array[i].setName("r" + itos(i + 1));
  }

  return array;
}

main(int argc, char * argv[])
{
  ClSimplexSolver xSolver;

  xSolver.reset();
  xSolver.setAutosolve(false);
  xSolver.setExplaining(true);

  if (argc != 2) {
    cerr << "Usage: " << argv[0] << " <filename>" << endl;
    exit(1);
  }

  ifstream in(argv[1]);

  int numVars = 0;
  ClVariable * vars;
  string action;
  char line[1000];

  in >> numVars;
  if (numVars <= 0) 
    {
	 cerr << "Error, non-positive num vars: " << numVars << endl;
	 exit(1);
    }

  vars = createVars(numVars);

  while (in >> action && action != "quit")
    {
	 in.getline(line, 1000);
	 if (action == "add")
	   {
		in.getline(line, 1000);
		istrstream strin(line);
		ClConstraint * cn = parseConstraint(strin, vars);
		
		try 
		  {
		    cout << "Adding " << *cn << endl;
		    xSolver.addConstraint(*cn);
		    cout << "Success" << endl;
		  }
		catch (const ExCLRequiredFailureWithExplanation &e) 
		  {
		    if (e.explanation().size() == 0) {
			 cout << "All constraints culpable" << endl;
		    }
		    else {
			 ClConstraintSet::const_iterator i;
			 for (i = e.explanation().begin(); 
				 i != e.explanation().end(); i++)
			   {
				cout << *(*i) << endl;
			   }
		    }
		    cout << "Failure" << endl;
		  }
	   }
	 else if (action == "print")
	   {
		cout << xSolver << endl;
	   }
	 else if (action == "solve")
	   {
		xSolver.solve();
		cout << "Solution: " << endl;
		for (int i = 0; i < numVars; i++) 
		  cout << i + 1 << " " << vars[i].value() << endl;
		cout << "Done." << endl;
	   }
	 else
	   {
		cerr << "Unknown action: " << action << " continuing at next line"
			<< endl;
	   }
    }
}
